/*
 * $Id: cia1.c,v 1.11 1996/04/01 06:00:00 ettore Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Vesa-Matti Puro  (vmp@lut.fi)
 *   Jarkko Sonninen  (sonninen@lut.fi)
 *   Jouko Valta      (jopi@stekt.oulu.fi)
 *
 * Timer fixes by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 * 
 * $Log: cia1.c,v $
 * Revision 1.11  1996/04/01 06:00:00  ettore
 * Bug fixes: timer & interrupts.
 *
 * Revision 1.10  1995/11/07  16:51:00  jopi
 * *** empty log message ***
 *
 * Revision 1.9  1995/06/28  19:40:56  jopi
 * VIA 6522 emulation
 * Control Bit Masks
 * Variable keyboard matrix sizes.
 *
 * Revision 1.8  1994/12/12  16:59:44  jopi
 * Defines for registers
 * C128 serial (slow mode) loop-back
 *
 * Revision 1.7  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.6  1994/02/18  16:10:57  jopi
 * New debugger & arg parser and x128
 *
 * Revision 1.5  1994/01/26  16:08:37  jopi
 * X64 version 0.2 PL 1
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:38:13  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  1993/06/13  08:19:36  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <time.h>

#include "vmachine.h"
#include "extern.h"
#include "cia.h"


/* #define CIA1_TIMER_DEBUG */

/* global */

BYTE    cia1[16];
BYTE    keyarr[KBD_ROWS];
int     cia1ta_stop = 0; /* maybe 1? */
int     cia1tb_stop = 0; /* maybe 1? */
int     cia1ta_interrupt = 0;
int     cia1tb_interrupt = 0;

/* local functions */

static void update_cia1ta ( int );
static void update_cia1tb ( int );
#ifndef VIA6522
static void update_tod ( void );
#endif

/*
 * Local variables
 */

static int   cia1int;	/* Interrupt Flag register for cia 1 */

static int cia1ta;	/* value of cia 1 timer A at last update */
static int cia1tb;	/* value of cia 1 timer B at last update */

static CLOCK cia1tau;	/* time when cia 1 timer A is updated */
static CLOCK cia1tbu;	/* time when cia 1 timer B is updated */

#ifndef VIA6522
static char   tod1virt = 0;	/* Virtual or Real-Time clock */

static char   tod1mode = 0;	/* Stopped */
static char   tod1freq;		/* 50/60 Hz select for cia 1 TOD */
static long   tod1cycles = 0;	/* TOD Virtual Mode displacemant */
static time_t tod1rel;		/* TOD 1 difference to system clock */
static time_t tod1alarm;	/* TOD 1 alarm time */

#endif


/* ------------------------------------------------------------------------- */
/* CIA 1 */


void    init_cia1(void)
{
#ifdef VIA6522
    cia1[P_AUX] = 0x40;  /* timer T1 run mode */
#endif
}


void    store_cia1(ADDRESS addr, BYTE byte)
{
    if (debugflg)
	printf("store cia1[%x] %x\n", (int) addr, (int) byte);


    switch (addr) {

      case P_PORT_A: /* port A */
#ifdef VIA6522
      case P_PORT_A_NHS: /* port A, no handshake */
#endif
	cia1[P_PORT_A] = (cia1[P_PORT_A] & ~cia1[P_DDIR_A]) |
	    (byte & cia1[P_DDIR_A]);
	break;

      case P_PORT_B: /* port B */
	cia1[P_PORT_B] = (cia1[P_PORT_B] & ~cia1[P_DDIR_B]) |
	    (byte & cia1[P_DDIR_B]);
	break;


	/* Timers */

      case P_TIMER_AH: /* Write timer A high */
	update_cia1ta(1);
#ifdef VIA6522
      case P_TIMER_L_AH: /* Write timer A high order latch */
#endif
	/* Clear T1 interrupt */
	cia1[addr] = byte;
	break;

      case P_TIMER_BH: /* Write timer B high */
	cia1[addr] = byte;
	/* Clear T2 interrupt */
	update_cia1tb(1);
	break;

	/*
	 * TOD clock is stopped by writing Hours, and restarted
	 * upon writing Tenths of Seconds.
	 *
	 * REAL:  TOD register + (wallclock - tod1rel)
	 * VIRT:  TOD register + (cycles - begin)/cycles_per_sec
	 */
#ifndef VIA6522
      case P_TOD_TEN: /* Time Of Day clock 1/10 s */
	if (P_CONT_B & 0x80) {  /* Set alarm */
	    tod1alarm = time(NULL);
	}
	else {
	    tod1mode = 1;  /* Run */
	    time(&tod1rel); /* seconds, use hrtime for better resolution */
	    tod1cycles = clk;

	    if (debugflg)
		printf("cia1 TOD set. time: ------  offset: %ld.\n", tod1rel);
	}
	cia1[addr] = byte;
	break;

      case P_TOD_HR:  /* Time Of Day clock hour */
	if (!(P_CONT_B & 0x80))  /* Set TOD Hours */
	    tod1mode = 0;  /* Stop */

      case P_TOD_SEC: /* Time Of Day clock sec */
      case P_TOD_MIN: /* Time Of Day clock min */

	if (P_CONT_B & 0x80) {  /* Set alarm */
	}
	else
	    cia1[addr] = byte;
	break;
#endif

      case P_SERIAL: /* Serial Port output buffer */
	cia1[addr] = byte;
	break;


	/* Interrupts */

#ifdef VIA6522
      case P_INTFLAG: /* 6522 Interrupt Flag Register */
	cia1int &= ~byte;
	break;
#endif

      case P_INTENAB: /* Interrupt Control Register */
#if defined (CIA1_TIMER_DEBUG)
        printf ("Cia#1 set P_INTENAB: 0x%x\n", byte);
#endif
	if (byte & IM_SET) {
	    /* set interrupts */
	    if (byte & IM_TA) {
	         cia1ta_interrupt = 1;
#if defined (CIA1_TIMER_DEBUG)
	         printf ("Cia#1 Timer A interrupt enabling requested\n");
#endif
	    }
	    if (byte & IM_TB) {
	         cia1tb_interrupt = 1;
#if defined (CIA1_TIMER_DEBUG)
	         printf ("Cia#1 Timer B interrupt enabling requested\n");
#endif
	    }
	    if ((byte & IM_TA) && !cia1ta_stop) { /* EP */
		update_cia1ta(0);
		int_clk[I_CIA1TA] = cia1tau + cia1ta;
#if defined (CIA1_TIMER_DEBUG)
		printf ("Cia#1 Timer A interrupt enabled\n");
#endif
	    }
	    if ((byte & IM_TB) && !cia1tb_stop) { /* EP */
		update_cia1tb(0);
		int_clk[I_CIA1TB] = cia1tbu + cia1tb;
#if defined (CIA1_TIMER_DEBUG)
		printf ("Cia#1 Timer B interrupt enabled\n");
#endif
	    }
#ifndef VIA6522
	    if (byte & IM_TOD)
		int_clk[I_CIA1TOD] = 0;	/* tod not implemented */
#endif
	    if (byte & IM_SP) {
		int_clk[I_CIA1SP] = 0;	/* serial data empty/full not impl. */
	    }

#ifdef VIA6522
	    if (byte & (IM_CB1 | IM_CB2 | IM_CA1 |IM_CA2))
#else
	    if (byte & IM_FLG)
#endif
		int_clk[I_CIA1FL] = 0;	/* FLAG1 not impl. */

	    cia1[addr] |= byte & 0x7f;
	}
	else {
	    /* clear interrupts */

	    if (byte & IM_TA) {
		int_clk[I_CIA1TA] = 0;
#if defined (CIA1_TIMER_DEBUG)
		printf ("Cia#1 Timer A interrupt disabled\n");
#endif
		cia1ta_interrupt = 0;
		/* cia1ta_stop =1; */
            }
	    if (byte & IM_TB) {
		int_clk[I_CIA1TB] = 0;
#if defined (CIA1_TIMER_DEBUG)
	        printf ("Cia#1 Timer B interrupt disabled\n");
#endif
		cia1tb_interrupt = 0;
		/* cia1tb_stop = 1; */
            }
#ifndef VIA6522
	    if (byte & IM_TOD)
		int_clk[I_CIA1TOD] = 0;
#endif
	    if (byte & IM_SP)
		int_clk[I_CIA1SP] = 0;
#ifdef VIA6522
	    if (byte & (IM_CB1 | IM_CB2 | IM_CA1 | IM_CA2))  /* only 1 used */
#else
	    if (byte & IM_FLG)
#endif
		int_clk[I_CIA1FL] = 0;

	    cia1[addr] &= ~byte;
	}

	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
	break;


	/* Control */

#ifdef VIA6522
      case P_AUX:
	cia1[addr] = byte;

	/* bit 7 timer 1 output to PB7 */
	/* bit 6 timer 1 run mode -- default seems to be continuous */

	/* bit 5 timer 2 count mode */
	if (byte & 32) {
	    update_cia1tb(0);	/* stop timer if mode == 1 */
	}

	/* bit 4, 3, 2 shift register control */

	/* bit 1, 0  latch enable port B and A */
	break;

      case P_PERIPH:
	cia1[addr] = byte;

	/* bit 7, 6, 5  CB2 handshake/interrupt control */
	/* bit 4  CB1 interrupt control */

	/* bit 3, 2, 1  CA2 handshake/interrupt control */
	/* bit 0  CA1 interrupt control */

	break;

#else  /* CIA 6526 */
      case P_CONT_A: /* control register A */
	cia1[addr] = byte;

	/* bit 7 tod frequency */
	tod1freq = (byte & 0x80) ? 50 : 60;

	/* bit 6 serial port mode */

	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia1ta(1);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */

	/* bit 0 start/stop timer */
	/* bit 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
#if defined (CIA1_TIMER_DEBUG)
	    printf ("Cia#1 Timer A stopped\n");
#endif
	    /* stop timer */
	    /* update_cia1ta(0); */
	    int_clk[I_CIA1TA] = 0;
	    cia1ta_stop = 1;
	} else {
#if defined (CIA1_TIMER_DEBUG)
	    printf ("Cia#1 Timer A restarted\n");
#endif
	    cia1ta_stop = 0;
	    if (cia1ta_interrupt) {
	        update_cia1ta(0);
	        int_clk[I_CIA1TA] = cia1tau + cia1ta;
	    }
	}
	break;

      case P_CONT_B: /* control register B */
	cia1[addr] = byte;
	/* bit 7 set alarm/tod clock */
	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia1tb(1);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */
	/* bit 0 start/stop timer */
	/* bit 6, 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    /* stop timer if mode = 01 or 11 */
	    int_clk[I_CIA1TB] = 0;           /* EP */
	    cia1tb_stop = 1;
#if defined (CIA1_TIMER_DEBUG)
	    printf ("Cia#1 Timer B stopped\n");
#endif
	    /* update_cia1tb(0); */
	} else {
#if defined (CIA1_TIMER_DEBUG)
	    printf ("Cia#1 Timer B restarted\n");
#endif
	    cia1tb_stop = 0;                
	    if (cia1tb_interrupt) {
	        update_cia1tb(0);               /* EP */
	        int_clk[I_CIA1TB] = cia1tbu + cia1tb;
	    }
	}
	break;
#endif  /* VIA6522 */

      default:
	cia1[addr] = byte;

    }  /* switch */
}


/* ------------------------------------------------------------------------- */

BYTE    read_cia1(ADDRESS addr)
{
    if (debugflg)
	printf("read cia1[%d]\n", addr);

    switch (addr) {

      case P_PORT_A: /* port A */
#ifdef VIA6522
      case P_PORT_A_NHS: /* port A, no handshake */
#endif
	{
	    int     t, i;
	    BYTE    j, v = 1;
	    t = ~(cia1[P_PORT_B] & cia1[P_DDIR_B]);
#ifdef C128
	    t |= ((read_vic(0x2f) & 7) >> 8);
#endif
	    for (j = i = 0; i < KBD_ROWS; v <<= 1, i++)
		if (t & keyarr[i])
		    j |= v;
	    j = ~j;

	    if (debugflg) {
		printf("read port A %d\n", j);
		printf("a: %x b:%x  ca: %x cb: %x joy: %x\n",
		       (int) j, (int) cia1[P_PORT_B],
		       (int) cia1[P_DDIR_A], (int) cia1[P_DDIR_B], joy[2]);
	    }
	    return (((j & ~cia1[P_DDIR_A]) | (cia1[P_PORT_A] & cia1[P_DDIR_A])) & ~joy[2]);
	}

      case P_PORT_B: /* port B */
	{
	    int     t, i;
	    BYTE    j;
	    t = cia1[P_PORT_A] & cia1[P_DDIR_A];  /* intr() updates keyboard */

	    for (j = i = 0; i < KBD_ROWS; t >>= 1, i++)
		if (!(t & 1))
		    j |= keyarr[i];
	    j = ~j;

	    if (debugflg) {
		printf("read port B %d\n", j);
		printf("a: %x b:%x  ca: %x cb: %x joy: %x\n",
		       (int) cia1[P_PORT_A], (int) j,
		       (int) cia1[P_DDIR_A], (int) cia1[P_DDIR_B], joy[1]);
	    }
	    return (((j & ~cia1[P_DDIR_B]) | (cia1[P_PORT_B] & cia1[P_DDIR_B])) & ~joy[1]);
	}


	/* Timers */

      case P_TIMER_AL: /* timer A low */
	update_cia1ta(0);
	return (cia1ta & 0xff);

      case P_TIMER_AH: /* timer A high */
	update_cia1ta(0);
	return ((cia1ta >> 8) & 0xff);

      case P_TIMER_BL: /* timer B low */
	update_cia1tb(0);
	return (cia1tb & 0xff);

      case P_TIMER_BH: /* timer B high */
	update_cia1tb(0);
	return ((cia1tb >> 8) & 0xff);

	/*
	 * TOD clock is lached by reading Hours, and released
	 * upon reading Tenths of Seconds. The counter itself
	 * keeps ticking all the time.
	 * Also note that this laching is different from the input one.
	 */
#ifndef VIA6522
      case P_TOD_TEN: /* Time Of Day clock 1/10 s */
	if (tod1mode == 1)  /* Running */
	    update_tod();
	    tod1mode = 1;  /* Run */
	return (cia1[addr]);

      case P_TOD_SEC: /* Time Of Day clock sec */
      case P_TOD_MIN: /* Time Of Day clock min */
	if (tod1mode == 1)  /* Running */
	    update_tod();
	return (cia1[addr]);

      case P_TOD_HR:  /* Time Of Day clock hour */
	if (tod1mode == 1)  /* Running */
	    update_tod();

	    tod1mode = 2;  /* Latched */
	return (cia1[addr]);
#endif

      case P_SERIAL: /* Serial Port Shift Register */
	return (cia1[addr]);


	/* Interrupts */

      case P_INTFLAG: /* Interrupt Flag Register */
	{
	    BYTE    t = cia1int | 8;   /****  This is ugly trick for x128 ****/
#ifdef VIA6522
	    if (cia1int & cia1[P_INTENAB])
		t |= 0x80;
#else
	    cia1int = 0;
#endif
	    return (t);
	}

#ifdef VIA6522
      case P_INTENAB: /* 6522 Interrupt Control Register */
	    return (cia1[P_INTENAB] | 0x80);
#endif

    }  /* switch */

    return (cia1[addr]);
}


/* ------------------------------------------------------------------------- */

void    int_cia1ta(void)
{
    if (debugflg)
	printf("cia1 timer A interrupt\n");

#ifdef VIA6522
    if (!(cia1[P_AUX] & 0x40)) { /* one-shot mode */
#else
    if (cia1[P_CONT_A] & 8) {	/* one-shot mode */
#endif
#if defined (CIA1_TIMER_DEBUG)
        printf ("Cia#1 Timer A interrupt -- one-shot mode: next int won't happen\n");
#endif
	cia1ta = 0;
	cia1tau = clk;
	int_clk[I_CIA1TA] = 0;
    }
    else {		/* continuous mode */
	update_cia1ta(1);
	int_clk[I_CIA1TA] = cia1tau + cia1ta;
    }
    cia1int = 0x80 | 1;

    /* count timer A underflow pulses */
    if ((cia1[15] & 0x61) == 0x41)
	if (--cia1tb <= 0)
	    int_clk[I_CIA1TB] = clk;
}


/*
 * Timer B can run in 2 (4) modes
 * cia1[f] & 0x60 == 0x00   count system 02 pulses
 * cia1[f] & 0x60 == 0x40   count timer A underflows
 * cia1[f] & 0x60 == 0x20 | 0x60 count CNT pulses => counter stops
 */


void    int_cia1tb(void)
{
    if (debugflg)
	printf("cia1 timer B interrupt\n");

#ifdef VIA6522
    if (!(cia1[P_AUX] & 0x20)) { /* one-shot mode */
#else
    if (cia1[P_CONT_B] & 8) {	/* one-shot mode */
#endif
#if defined (CIA1_TIMER_DEBUG)
        printf ("Cia#1 Timer B interrupt -- one-shot mode: next int won't happen\n");
#endif
	cia1tb = 0;
	cia1tbu = clk;
	int_clk[I_CIA1TB] = 0;
    }
    else {		/* continuous mode */
	update_cia1tb(1);
	if ((cia1[15] & 0x61) == 0x01)
	    int_clk[I_CIA1TB] = cia1tbu + cia1tb;
    }
    cia1int = 0x80 | 2;
}


/* ------------------------------------------------------------------------- */

static void update_cia1ta(int force)
{
    if (force)
	cia1ta = cia1[P_TIMER_AL] + (cia1[P_TIMER_AH] << 8);
    else if ((cia1[14] & 0x21) == 1)
	cia1ta = (int)(clk - cia1tau + cia1ta) & 0xffff;
    cia1tau = clk;
}


static void update_cia1tb(int force)
{
    if (force) {
	cia1tb = cia1[P_TIMER_BL] + (cia1[P_TIMER_BH] << 8);
    }
    else if ((cia1[15] & 0x61) == 1)
	/* count system 02 pulses */
	cia1tb = (int)(clk - cia1tbu + cia1tb) & 0xffff;
    cia1tbu = clk;
}


#ifndef VIA6522
static void update_tod(void)
{
    long   c;
    time_t t = 0;

    /*if (tod1mode != 1) return; */  /* Not Running */

    if (tod1virt) {
	c = clk - tod1cycles;
	t = (double)c / (tod1freq ? 9852.48 : 10227.30); /* CYCLES / 0.01 SEC */
    }
    else {
	time(&t);
	t -= tod1rel;
    }

    cia1[P_TOD_TEN] = 0;			/* Time Of Day clock 1/10 s */
    cia1[P_TOD_SEC] = (BYTE)(t/3600);		/* Time Of Day clock sec */
    cia1[P_TOD_MIN] = (BYTE)((t%3600)/60);	/* Time Of Day clock min */
    cia1[P_TOD_HR]  = (BYTE)(t%60);
}
#endif


int     show_keyarr(void)
{
    int     i, j;
    for (j = 0; j < KBD_ROWS; j++) {
	printf("%d:", j);
	for (i = 0x80; i; i >>= 1)
	    printf(" %d", keyarr[j] & i);
	printf("\n");
    }
    return (0);
}
