/*
 * $Id: cia2.c,v 1.9 1996/04/01 06:02:00 ettore Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Vesa-Matti Puro  (vmp@lut.fi)
 *   Jarkko Sonninen  (sonninen@lut.fi)
 *   Jouko Valta      (jopi@stekt.oulu.fi)
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 *
 * $Log: cia2.c,v $
 * Revision 1.9  1996/04/01 06:02:00  ettore
 * Bug fixes: timer & interrupts.
 *
 * Revision 1.8  1995/11/07  16:51:00  jopi
 * *** empty log message ***
 *
 * Revision 1.7  1995/06/28  19:40:56  jopi
 * VIA 6522 emulation
 * Control Bit Masks
 *
 * Revision 1.6  1994/12/12  16:59:44  jopi
 * New VIC control
 *
 * Revision 1.5  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:38:16  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  1993/06/13  08:19:36  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <time.h>

#include "vmachine.h"
#include "extern.h"
#include "cia.h"

/* #define CIA2_TIMER_DEBUG */

/* global */

BYTE    cia2[16];
int     cia2ta_stop = 1;
int     cia2tb_stop = 1;
int	cia2ta_interrupt = 0;
int	cia2tb_interrupt = 0;

/* local functions */

static void update_cia2ta ( int );
static void update_cia2tb ( int );
#ifndef VIA6522
static void update_tod ( void );
#endif


/*
 * Local variables
 */

static int cia2int;	/* Interrupt Flag register for cia 2 */

static int cia2ta;	/* value of cia 2 timer A at last update */
static int cia2tb;	/* value of cia 2 timer B at last update */

static CLOCK cia2tau;	/* time when cia 2 timer A is updated */
static CLOCK cia2tbu;	/* time when cia 2 timer B is updated */

#ifndef VIA6522
static char   tod2virt = 0;	/* Virtual or Real-Time clock */

static char   tod2mode = 0;	/* Stopped */
static char   tod2freq;		/* 50/60 Hz select for cia 2 TOD */
static long   tod2cycles = 0;	/* TOD Virtual Mode displacemant */
static time_t tod2rel;		/* TOD 2 difference to system clock */
static time_t tod2alarm;	/* TOD 2 alarm time */

extern int vbank;	/* VIC bank */
#endif


/* ------------------------------------------------------------------------- */
/* CIA 2 */


void    init_cia2(void)
{
#ifdef VIA6522
    cia1[P_AUX] = 0x40;	/* timer T1 run mode */
#endif
}


void    store_cia2(ADDRESS addr, BYTE byte)
{
    if (debugflg)
	printf("store cia2[%x] %x\n", (int) addr, (int) byte);

    switch (addr) {

      case P_PORT_A: /* port A */
#ifdef VIA6522
      case P_PORT_A_NHS: /* port A, no handshake */
#endif

#ifndef VIA6522  /* C64 CIA 2 */
	{
	    BYTE    tmp = cia2[P_PORT_A];
#endif
	    cia2[P_PORT_A] = (cia2[P_PORT_A] & ~cia2[P_DDIR_A]) |
		(byte & cia2[P_DDIR_A]);
#ifndef VIA6522
#ifdef VIC_II 
	    vbank = ((cia2[0] & 3) ^ 3) << 14;
#endif
	    if ((tmp ^ cia2[0]) & 3)
		screen_init();
	}
#endif  /* C64 CIA 2 */
	break;

      case P_PORT_B: /* port B */
	cia2[P_PORT_B] = (cia2[P_PORT_B] & ~cia2[P_DDIR_B]) |
	    (byte & cia2[P_DDIR_B]);
	break;


	/* Timers */

      case P_TIMER_AH: /* Write timer A high */
	update_cia2ta(1);
#ifdef VIA6522
      case P_TIMER_L_AH: /* Write timer A high order latch */
#endif
	/* Clear T1 interrupt */
	cia2[addr] = byte;
	break;

      case P_TIMER_BH: /* Write timer B high */
	cia2[addr] = byte;
	/* Clear T2 interrupt */
	update_cia2tb(1);
	break;

	/*
	 * TOD clock is stopped by writing Hours, and restarted
	 * upon writing Tenths of Seconds.
	 */
#ifndef VIA6522
      case P_TOD_TEN: /* Time Of Day clock 1/10 s */
	if (P_CONT_B & 0x80) {  /* Set alarm */
	    tod2alarm = (int)time(NULL);
	}
	else {
	    tod2mode = 1;  /* Run */
	    time(&tod2rel); /* seconds, use hrtime for better resolution */
	    tod2cycles = clk;

	    if (debugflg)
		printf("cia2 TOD set. time: ------  offset: %ld.\n", tod2rel);
	}
	cia2[addr] = byte;
	break;

      case P_TOD_HR:  /* Time Of Day clock hour */
	if (!(P_CONT_B & 0x80))  /* Set TOD Hours */
	    tod2mode = 0;  /* Stop */

      case P_TOD_SEC: /* Time Of Day clock sec */
      case P_TOD_MIN: /* Time Of Day clock min */

	if (P_CONT_B & 0x80) {  /* Set alarm */
	}
	else
	    cia2[addr] = byte;
	break;
#endif

      case P_SERIAL: /* Serial Port output buffer */
	cia2[addr] = byte;
	break;


	/* Interrupts */

#ifdef VIA6522
      case P_INTFLAG: /* 6522 Interrupt Flag Register */
	cia2int &= ~byte;
	break;
#endif

      case P_INTENAB: /* Interrupt Control Register */
#if defined (CIA2_TIMER_DEBUG)
        printf ("Cia#2 set P_INTENAB: 0x%x\n", byte);
#endif
	if (byte & IM_SET) {
	    /* set interrupts */
	    if (byte & IM_TA) {
		cia2ta_interrupt = 1;
#if defined (CIA2_TIMER_DEBUG)
	        printf ("Cia#2 Timer A interrupt enabling requested\n");
#endif
	    }
	    if (byte & IM_TB) {
	        cia2ta_interrupt = 1;
#if defined (CIA2_TIMER_DEBUG)
	        printf ("Cia#2 Timer B interrupt enabling requested\n");
#endif
	    }
	    if ((byte & IM_TA) && !cia2ta_stop) { /* EP */
		update_cia2ta(0);
		int_clk[I_CIA2TA] = cia2tau + cia2ta;
#if defined (CIA2_TIMER_DEBUG)
		printf ("Cia#2 Timer A interrupt enabled\n");
#endif
	    }
	    if ((byte & IM_TB) && !cia2tb_stop) { /* EP */
		update_cia2tb(0);
		int_clk[I_CIA2TB] = cia2tbu + cia2tb;
#if defined (CIA2_TIMER_DEBUG)
		printf ("Cia#2 Timer B interrupt enabled\n");
#endif
	    }
#ifndef VIA6522
	    if (byte & IM_TOD)
		int_clk[I_CIA2TOD] = 0;	/* tod not implemented */
#endif
	    if (byte & IM_SP) {
		int_clk[I_CIA2SP] = 0;	/* serial data empty/full not impl. */
	    }

#ifdef VIA6522
	    if (byte & (IM_CB1 | IM_CB2 | IM_CA1 |IM_CA2))
#else
	    if (byte & IM_FLG)
#endif
		int_clk[I_CIA2FL] = 0;	/* FLAG1/ RESTORE key */

	    cia2[addr] |= byte & 0x7f;
	}
	else {
	    /* clear interrupts */

	    if (byte & IM_TA) {
		int_clk[I_CIA2TA] = 0;
#if defined (CIA2_TIMER_DEBUG)
		printf ("Cia#2 Timer A interrupt disabled\n");
#endif
		cia2ta_interrupt = 0;
		/* cia1ta_stop = 1; */
    	    }
	    if (byte & IM_TB) {
		int_clk[I_CIA2TB] = 0;
#if defined (CIA2_TIMER_DEBUG)
	        printf ("Cia#2 Timer B interrupt disabled\n");
#endif
		cia2tb_interrupt = 0;
	        /* cia2tb_stop = 1; */
	    }
#ifndef VIA6522
	    if (byte & IM_TOD)
		int_clk[I_CIA2TOD] = 0;
#endif
	    if (byte & IM_SP)
		int_clk[I_CIA2SP] = 0;
#ifdef VIA6522
	    if (byte & (IM_CB1 | IM_CB2 | IM_CA1 | IM_CA2))  /* only 1 used */
#else
	    if (byte & IM_FLG)
#endif
		int_clk[I_CIA2FL] = 0;

	    cia2[addr] &= ~byte;
	}

	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
	break;


	/* Control */

#ifdef VIA6522
      case P_AUX:
	cia2[addr] = byte;

	/* bit 7 timer 1 output to PB7 */
	/* bit 6 timer 1 run mode -- default seems to be continuous */

	/* bit 5 timer 2 count mode */
	if (byte & 32) {
	    update_cia2tb(0);	/* stop timer if mode == 1 */
	}

	/* bit 4, 3, 2 shift register control */

	/* bit 1, 0  latch enable port B and A */
	break;

      case P_PERIPH:
	cia2[addr] = byte;

	/* bit 7, 6, 5  CB2 handshake/interrupt control */
	/* bit 4  CB1 interrupt control */

	/* bit 3, 2, 1  CA2 handshake/interrupt control */
	/* bit 0  CA1 interrupt control */

	break;

#else  /* CIA 6526 */
      case P_CONT_A: /* control register A */
	cia2[addr] = byte;

	/* bit 7 tod frequency */
	tod2freq = (byte & 0x80) ? 50 : 60;

	/* bit 6 serial port mode */

	/* bit 4 force load */
	if (byte & 0x10)
	    update_cia2ta(0);

	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */

	/* bit 0 start/stop timer */
	/* bit 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    /* stop timer */
#if defined (CIA2_TIMER_DEBUG)
	    printf ("Cia#2 Timer A stopped\n");
#endif
	    /* update_cia2ta(0); */  
	    int_clk[I_CIA2TA] = 0;
	    cia2ta_stop = 1;
	} else {
#if defined (CIA2_TIMER_DEBUG)
	    printf ("Cia#2 Timer A restarted\n");
#endif
	    cia2ta_stop = 0;
	    if (cia2ta_interrupt) {
	        update_cia2ta(0);
	        int_clk[I_CIA2TA] = cia2tau + cia2ta;
	    }
	}
	break;

      case P_CONT_B: /* control register B */
	cia2[addr] = byte;
	/* bit 7 set alarm/tod clock */
	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia2tb(1);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */
	/* bit 0 start/stop timer */
	/* bit 6, 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    /* stop timer if mode = 01 or 11 */
	    int_clk[I_CIA2TB] = 0;
	    cia2tb_stop = 1;
#if defined (CIA2_TIMER_DEBUG)
	    printf ("Cia#2 Timer B stopped\n");
#endif
	    /* update_cia2tb(0); */
	} else {
#if defined (CIA2_TIMER_DEBUG)
	    printf ("Cia#2 Timer B restarted\n");
#endif
	    cia2tb_stop = 0;
	    if (cia2tb_interrupt) {
	        update_cia2tb(0);	/* EP */
	        int_clk[I_CIA2TB] = cia2tbu + cia2tb;
	    }
	}
	break;
#endif  /* VIA6522 */

      default:
	cia2[addr] = byte;

    }  /* switch */
}


/* ------------------------------------------------------------------------- */

BYTE    read_cia2(ADDRESS addr)
{
    switch (addr) {

      case P_PORT_A: /* port A */
#ifdef VIA6522
      case P_PORT_A_NHS: /* port A, no handshake */
#endif
	return ((cia2[P_PORT_A] & cia2[P_DDIR_A]) | (0x7f & ~cia2[P_DDIR_A]));

      case P_PORT_B: /* port B */
	return ((cia2[P_PORT_B] & cia2[P_DDIR_B]) | (0xff & ~cia2[P_DDIR_B]));


	/* Timers */

      case P_TIMER_AL: /* timer A low */
	update_cia2ta(0);
	return (cia2ta & 0xff);

      case P_TIMER_AH: /* timer A high */
	update_cia2ta(0);
	return ((cia2ta >> 8) & 0xff);

      case P_TIMER_BL: /* timer B low */
	update_cia2tb(0);
	return (cia2tb & 0xff);

      case P_TIMER_BH: /* timer B high */
	update_cia2tb(0);
	return ((cia2tb >> 8) & 0xff);

	/*
	 * TOD clock is lached by reading Hours, and released
	 * upon reading Tenths of Seconds. The counter itself
	 * keeps ticking all the time.
	 * Also note that this laching is different from the input one.
	 */
#ifndef VIA6522
      case P_TOD_TEN: /* Time Of Day clock 1/10 s */
	if (tod2mode == 1)  /* Running */
	    update_tod();
	    tod2mode = 1;  /* Run */
	return (cia2[addr]);

      case P_TOD_SEC: /* Time Of Day clock sec */
      case P_TOD_MIN: /* Time Of Day clock min */
	if (tod2mode == 1)  /* Running */
	    update_tod();
	return (cia2[addr]);

      case P_TOD_HR:  /* Time Of Day clock hour */
	if (tod2mode == 1)  /* Running */
	    update_tod();

	    tod2mode = 2;  /* Latched */
	return (cia2[addr]);
#endif

      case P_SERIAL: /* Serial Port Shift Register */
	return (cia2[addr]);


	/* Interrupts */

      case P_INTFLAG: /* Interrupt Flag Register */
	{
	    BYTE    t = cia2int;
#ifdef VIA6522
	    if (cia2int & cia2[P_INTENAB])
		t |= 0x80;
#else
	    cia2int = 0;
#endif
	    return (t);
	}

#ifdef VIA6522
      case P_INTENAB: /* 6522 Interrupt Control Register */
	    return (cia2[P_INTENAB] | 0x80);
#endif

    }  /* switch */

    return (cia2[addr]);
}


/* ------------------------------------------------------------------------- */

void    int_cia2ta(void)
{
    if (debugflg)
	printf("cia2 timer A interrupt\n");

#ifdef VIA6522
    if (!(cia2[P_AUX] & 0x40)) { /* one-shot mode */
#else
    if (cia2[P_CONT_A] & 8) {	/* one-shot mode */
#endif
#if defined (CIA2_TIMER_DEBUG)
        printf ("Cia#2 Timer A interrupt -- one-shot mode: next int won't happen\n");
#endif
	cia2ta = 0;
	cia2tau = clk;
	int_clk[I_CIA2TA] = 0;
    }
    else {		/* continuous mode */
	update_cia2ta(1);
	int_clk[I_CIA2TA] = cia2tau + cia2ta;
    }
    cia2int = 0x80 | 1;

    /* count timer A underflow pulses */
    if ((cia2[15] & 0x61) == 0x41)
	if (--cia2tb <= 0)
	    int_clk[I_CIA2TB] = clk;
}


/*
 * Timer B can run in 2 (4) modes
 * cia2[f] & 0x60 == 0x00   count system 02 pulses
 * cia2[f] & 0x60 == 0x40   count timer A underflows
 * cia2[f] & 0x60 == 0x20 | 0x60 count CNT pulses => counter stops
 */


void    int_cia2tb(void)
{
    if (debugflg)
	printf("cia2 timer B interrupt\n");

#ifdef VIA6522
    if (!(cia2[P_AUX] & 0x20)) { /* one-shot mode */
#else
    if (cia2[P_CONT_B] & 8) {	/* one-shot mode */
#endif
#if defined (CIA1_TIMER_DEBUG)
        printf ("Cia#1 Timer B interrupt -- one-shot mode: next int won't happen\n");
#endif
	cia2tb = 0;
	cia2tbu = clk;
	int_clk[I_CIA2TB] = 0;
    }
    else {		/* continuous mode */
	update_cia2tb(1);
	if ((cia2[15] & 0x61) == 0x01)
	    int_clk[I_CIA2TB] = cia2tbu + cia2tb;
    }
    cia2int = 0x80 | 2;
}


/* ------------------------------------------------------------------------- */

static void update_cia2ta(int force)
{
    if (force)
	cia2ta = cia2[P_TIMER_AL] + (cia2[P_TIMER_AH] << 8);
    else if ((cia2[14] & 0x21) == 1)
	cia2ta = (int)(clk - cia2tau + cia2ta) & 0xffff;
    cia2tau = clk;
}


static void update_cia2tb(int force)
{
    if (force) {
	cia2tb = cia2[P_TIMER_BL] + (cia2[P_TIMER_BH] << 8);
    }
    else if ((cia2[15] & 0x61) == 1)
	/* count system 02 pulses */
	cia2tb = (int)(clk - cia2tbu + cia2tb) & 0xffff;
    cia2tbu = clk;
}


#ifndef VIA6522
static void update_tod(void)
{
    long   c;
    time_t t = 0;

    /*if (tod2mode != 1) return; */  /* Not Running */

    if (tod2virt) {
	c = clk - tod2cycles;
	t = (double)c / (tod2freq ? 9852.48 : 10227.30); /* CYCLES / 0.01 SEC */
    }
    else {
	time(&t);
	t -= tod2rel;
    }

    cia2[P_TOD_TEN] = 0;			/* Time Of Day clock 1/10 s */
    cia2[P_TOD_SEC] = (BYTE)(t/3600);		/* Time Of Day clock sec */
    cia2[P_TOD_MIN] = (BYTE)((t%3600)/60);	/* Time Of Day clock min */
    cia2[P_TOD_HR]  = (BYTE)(t%60);
}
#endif
