/*
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *   Andre Fachat     (a.fachat@physik.tu-chemnitz.de)
 *
 * 
 * Nearly completely rewritten. 1996/03/27 a.fachat
 *
 * moved it to via.c and removed all cia specific stuff
 * to have cleaner PET i/o handling	a.fachat
 * 
 * Revision 1.10  1995/11/07  16:51:00  jopi
 * *** empty log message ***
 *
 * Revision 1.9  1995/06/28  19:40:56  jopi
 * VIA 6522 emulation
 * Control Bit Masks
 * Variable keyboard matrix sizes.
 *
 * Revision 1.8  1994/12/12  16:59:44  jopi
 * Defines for registers
 * C128 serial (slow mode) loop-back
 *
 * Revision 1.7  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.6  1994/02/18  16:10:57  jopi
 * New debugger & arg parser and x128
 *
 * Revision 1.5  1994/01/26  16:08:37  jopi
 * X64 version 0.2 PL 1
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:38:13  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  1993/06/13  08:19:36  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <time.h>

#include "vmachine.h"
#include "extern.h"
#include "via.h"
#include "parallel.h"

#include "cpu.h"

/* global */

BYTE    via1[16];
BYTE    keyarr[KBD_ROWS];

/* local functions */

static int viadebug = 0;

static void update_via1ta ( int );
static void update_via1tb ( int );

/*
 * Local variables
 */

static int via1ifr;	/* Interrupt Flag register for via */
static int via1ier;	/* Interrupt Flag register for via */

static int via1ta;	/* value of via timer 1 at last update */
static int via1tb;	/* value of via timer 2 at last update */

static CLOCK via1tau;	/* time when via timer 1 is updated */
static CLOCK via1tbu;	/* time when via timer 2 is updated */


/* ------------------------------------------------------------------------- */
/* VIA 1 */


/* 
 * according to Rockwell, all internal registers are cleared, except
 * for the Timer (1 and 2, counter and latches) and the shift register.
 */ 

void    reset_via1(void)
{
    int i;
    if(viadebug) printf("VIA1: reset\n");   
    /* clear registers */ 
    for(i=0;i<4;i++) via1[i]=0;
    for(i=11;i<16;i++) via1[i]=0;

/* 
 * The following line doesn't appear in the specs...
 * If it's for the PET, it's useless. A PET uses the Rasterline 
 * IRQ.
 */

    /*via1[VIA_ACR] = 0x40;*/  /* timer T1 run mode */

    /* set IEC output lines */
    par_set_atn(0);
    par_set_nrfd(0);

    /* disable vice interrupts */
    int_clk[I_VIA1T1] = 0;
    int_clk[I_VIA1T2] = 0;
    int_clk[I_VIA1SR] = 0;
    int_clk[I_VIA1FL] = 0;

    via1ier = 0;
    via1ifr = 0;
}


void    store_via1(ADDRESS addr, BYTE byte)
{
    if (debugflg)
	printf("store via1[%x] %x\n", (int) addr, (int) byte);


    switch (addr) {

	/* these are done with saving the value */

      case VIA_PRA: /* port A */
      case VIA_PRA_NHS: /* port A, no handshake */
      case VIA_DDRA:
      case VIA_DDRB:
      case VIA_T2LL:
      case VIA_SR:
    	via1[addr] = byte;
	break;

	/* Port registers */

      case VIA_PRB: /* port B */
    	via1[addr] = byte;
	byte &= via1[VIA_DDRB];
	par_set_nrfd(!(byte & 0x02));
	par_set_atn(!(byte & 0x04));
	break;

	/* Timers */

      case VIA_T1CL:
      case VIA_T1LL:
	via1[VIA_T1LL] = byte;
	break;

      case VIA_T1CH: /* Write timer A high */
	if(pardebug) printf("Write timer A high: %d\n",byte);
    	via1[VIA_T1LH] = byte;
	/* load counter with latch value */
	via1[VIA_T1CL] = via1[VIA_T1LL];
	via1[VIA_T1CH] = via1[VIA_T1LH];
	/* Clear T1 interrupt */
	via1ifr &= ~VIA_IM_T1;
	update_via1ta(1);
	break;

      case VIA_T1LH: /* Write timer A high order latch */
	via1[addr] = byte;
	/* Clear T1 interrupt */
	via1ifr &= ~VIA_IM_T1;
	break;

      case VIA_T2CH: /* Write timer 2 high */
	via1[VIA_T2CH] = byte;
	via1[VIA_T2CL] = via1[VIA_T2LL]; /* bogus, both are identical */
	update_via1tb(1);
	/* Clear T2 interrupt */
	via1ifr &= ~VIA_IM_T2;
	break;

	/* Interrupts */

      case VIA_IFR: /* 6522 Interrupt Flag Register */
	via1ifr &= ~byte;
	break;

      case VIA_IER: /* Interrupt Control Register */
	if (byte & VIA_IM_IRQ) {
	    /* set interrupts */
/* These interrupts are all generated even without ier set! */
/*
	    if (byte & VIA_IM_T1) {
		update_via1ta(0);
		int_clk[I_VIA1T1] = via1tau + via1ta;
	    }
	    if (byte & VIA_IM_T2) {
		update_via1tb(0);
		int_clk[I_VIA1T2] = via1tbu + via1tb;
	    }
	    if (byte & VIA_IM_SR) {
		int_clk[I_VIA1SR] = 0;	*//* serial data empty/full not impl. *//*
	    }
	    if (byte & (VIA_IM_CB1 | VIA_IM_CB2 | VIA_IM_CA1 | VIA_IM_CA2))
		int_clk[I_VIA1FL] = 0;	*//* FLAG1 not impl. *//*
*/
	    via1ier |= byte & 0x7f;
	}
	else {
	    /* clear interrupts */
/*
	    if (byte & VIA_IM_T1)
		int_clk[I_VIA1T1] = 0;
	    if (byte & VIA_IM_T2)
		int_clk[I_VIA1T1] = 0;
	    if (byte & VIA_IM_SR)
		int_clk[I_VIA1SR] = 0;
	    if (byte & (VIA_IM_CB1 | VIA_IM_CB2 | VIA_IM_CA1 | VIA_IM_CA2)) 
		int_clk[I_VIA1FL] = 0;
*/
	    via1ier &= ~byte;
	}

	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
	break;

	/* Control */

      case VIA_ACR:
	if(viadebug) 
	  printf("VIA1: write %02x to ACR at PC=%04x, previous value = %02x\n",
			byte, PC, via1[VIA_ACR]);
    	via1[addr] = byte;
	/* bit 7 timer 1 output to PB7 */
	/* bit 6 timer 1 run mode -- default seems to be continuous */

	/* bit 5 timer 2 count mode */
	if (byte & 32) {
/* TODO */
/*	    update_via1tb(0);*/	/* stop timer if mode == 1 */
	}

	/* bit 4, 3, 2 shift register control */

	/* bit 1, 0  latch enable port B and A */
	break;

      case VIA_PCR:
    	via1[addr] = byte;
	if(viadebug) printf("VIA1: write %02x to PCR\n",byte);
	/* bit 7, 6, 5  CB2 handshake/interrupt control */
	/* bit 4  CB1 interrupt control */

	/* bit 3, 2, 1  CA2 handshake/interrupt control */
	/* bit 0  CA1 interrupt control */

	crtc_set_char( byte & 2 ); /* hack for switching PET charrom */
	break;

      default:
    	via1[addr] = byte;
	break;
    }  /* switch */
}


/* ------------------------------------------------------------------------- */

BYTE    read_via1(ADDRESS addr)
{
    if (debugflg)
	printf("read via1[%d]\n", addr);

    switch (addr) {

      case VIA_PRA: /* port A */
      case VIA_PRA_NHS: /* port A, no handshake */
	{
	    BYTE    j = 255;
	    /* nothing to read yet */
	    if (debugflg) {
		printf("read port A %d\n", j);
		printf("a: %x b:%x  ca: %x cb: %x joy: %x\n",
		       (int) j, (int) via1[VIA_PRB],
		       (int) via1[VIA_DDRA], (int) via1[VIA_DDRB], joy[2]);
	    }
	    return ((j & ~via1[VIA_DDRA]) | (via1[VIA_PRA] & via1[VIA_DDRA]));
	}

      case VIA_PRB: /* port B */
	{
	    BYTE    j;
	    /* read parallel IEC interface line states */
	    j = 255 - (par_nrfd ? 64:0) - (par_ndac ? 1:0) - (par_dav ? 128:0);
	    /* vertical retrace */
	    j -= crtc_offscreen() ? 32:0;
	    if (0 /*pardebug*/) {
		printf("read port B %d\n", j);
		printf("a: %x b:%x  ca: %x cb: %x joy: %x\n",
		       (int) via1[VIA_PRA], (int) j,
		       (int) via1[VIA_DDRA], (int) via1[VIA_DDRB], joy[1]);
	    }
	    return ((j & ~via1[VIA_DDRB]) | (via1[VIA_PRB] & via1[VIA_DDRB]));
	}

	/* Timers */

      case VIA_T1CL: /* timer A low */
/*	update_via1ta(0);*/
	via1ifr &= ~VIA_IM_T1;
	return ((via1ta - clk + via1tau) & 0xff);

      case VIA_T1CH: /* timer A high */
/*	update_via1ta(0);*/
	return (((via1ta - clk + via1tau) >> 8) & 0xff);

      case VIA_T2CL: /* timer B low */
/*	update_via1tb(0);*/
	via1ifr &= ~VIA_IM_T2;
	return ((via1tb - clk + via1tbu) & 0xff);

      case VIA_T2CH: /* timer B high */
/*	update_via1tb(0);*/
	return (((via1tb - clk + via1tbu) >> 8) & 0xff);

      case VIA_SR: /* Serial Port Shift Register */
	return (via1[addr]);

	/* Interrupts */

      case VIA_IFR: /* Interrupt Flag Register */
	{
	    BYTE    t = via1ifr;
	    if (via1ifr & via1[VIA_IER])
		t |= 0x80;
	    if(viadebug || pardebug) printf("Read Intflag VIA gives %x\n",t);
	    return (t);
	}

      case VIA_IER: /* 6522 Interrupt Control Register */
	    return (via1[VIA_IER] | 0x80);

      case VIA_ACR:
	    if(viadebug) printf("read ACR gives %02x\n",via1[addr]);
	    return via1[addr];
    }  /* switch */

    return (via1[addr]);
}


/* ------------------------------------------------------------------------- */

int    int_via1ta(void)
{
    if (debugflg || pardebug)
	printf("via1 timer A interrupt\n");

    if (!(via1[VIA_ACR] & 0x40)) { /* one-shot mode */
	if(viadebug || pardebug) printf("one-shot mode via1 timer A!\n");
	via1ta = 0;
	via1tau = clk;
	int_clk[I_VIA1T1] = 0;
    }
    else {		/* continuous mode */
	/* load counter with latch value */
	via1[VIA_T1CL] = via1[VIA_T1LL];
	via1[VIA_T1CH] = via1[VIA_T1LH];
	update_via1ta(1);
/*	int_clk[I_CIA1TA] = via1tau + via1ta;*/
    }
    via1ifr |= VIA_IM_T1; 
    return !(via1ier & VIA_IM_T1);
}


/*
 * Timer B is always in one-shot mode
 */


int    int_via1tb(void)
{
    if (debugflg)
	printf("via1 timer B interrupt\n");

    via1tb = 0;
    via1tbu = clk;
    int_clk[I_CIA1TB] = 0;

    via1ifr |= VIA_IM_T2;
    return !(via1ier & VIA_IM_T2);
}


/* ------------------------------------------------------------------------- */

static void update_via1ta(int force)
{
    if(force) {
       if(viadebug) 
	  printf("update via1 timer A : latch=%d, counter =%d, clk = %ld\n",
      		via1[VIA_T1CL] + (via1[VIA_T1CH] << 8),
      		via1[VIA_T1LL] + (via1[VIA_T1LH] << 8),
		clk);

      via1ta = via1[VIA_T1CL] + (via1[VIA_T1CH] << 8);
      via1tau = clk;
      int_clk[I_VIA1T1] = via1tau + via1ta;
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif  
    }
}


static void update_via1tb(int force)
{
    if(force) {
      via1tb = via1[VIA_T2CL] + (via1[VIA_T2CH] << 8);
      via1tbu = clk;
      int_clk[I_VIA1T2] = via1tbu + via1tb;
    	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
    }
}

int     show_keyarr(void)
{
    int     i, j;
    for (j = 0; j < KBD_ROWS; j++) {
        printf("%d:", j);
        for (i = 0x80; i; i >>= 1)
            printf(" %d", keyarr[j] & i);
        printf("\n");
    }
    return (0);
}


